#include "LoopSimplify.hpp"
#include "BasicBlock.hpp"
#include "Function.hpp"
#include "Instruction.hpp"

#include "LoopDetection.hpp"
#include "Value.hpp"
#include <algorithm>
#include <cassert>

using namespace std;
using PhiPair = pair<Value *, BasicBlock *>;

// 将phi指令的op按照来源是loop内外分离
pair<vector<PhiPair>, vector<PhiPair>> LoopSimplify::split_phi_op(PhiInst *phi,
                                                                  Loop &loop) {
    vector<PhiPair> inner, outer;
    for (auto &&[op, bb] : phi->get_pairs()) {
        auto loopbbs = loop.get_blocks();
        if (find(loopbbs.begin(), loopbbs.end(), bb) != loopbbs.end()) {
            inner.emplace_back(op, bb);
        } else {
            outer.emplace_back(op, bb);
        }
    }
    return {inner, outer};
}
/*Loop pre-header insertion guarantees that there is a single, non-critical
 * entry edge from outside of the loop to the loop header. This simplifies a
 * number of analyses and transformations, such as LICM.
 * 保证一个循环有一个preheader，且head中phi引用的外部变量都来自preheader*/
BasicBlock *LoopSimplify::preheader_insertion(Loop &loop) {
    auto header = loop.get_header();
    auto func = header->get_parent();
    auto preheader = BasicBlock::create(func->get_parent(), "", func);

    vector<PhiInst *> phis;
    for (auto &&inst : header->get_instructions()) {
        if (not inst.is_phi()) {
            break;
        }
        phis.push_back(dynamic_cast<PhiInst *>(&inst));
    }

    // modify phi inst
    for (auto &&phi : phis) {
        auto [inner, outer] = split_phi_op(phi, loop);
        if (outer.size() == 0) {
            continue;
        }
        if (inner.size() == 0) {
            preheader->remove_instr(phi);
            preheader->add_instruction(phi);
            continue;
        }
        if (outer.size() == 1) {
            inner.emplace_back(outer[0].first, preheader);
            // FIXME
            for (unsigned i = 0; i < phi->get_num_operand(); ++i)
                if (phi->get_operand(i))
                    phi->get_operand(i)->remove_use(phi, i);
            phi->remove_all_operands();

            for (auto [op, bb] : inner) {
                assert(op->get_type() == phi->get_type() and
                       "op_type must equal to phi_type");
                phi->add_operand(op);
                phi->add_operand(bb);
            }

            continue;
        }
        // create a new phi in preheader
        auto phi_outer = PhiInst::create_phi(phi->get_type(), preheader);
        preheader->add_instr_begin(phi_outer);

        // for (unsigned i = 0; i < phi_outer->get_num_operand(); ++i)
        //     if (phi_outer->get_operand(i))
        //         phi_outer->get_operand(i)->remove_use(phi_outer, i);
        // phi_outer->remove_all_operands();

        for (auto [op, bb] : outer) {
            assert(op->get_type() == phi_outer->get_type());
            phi_outer->add_operand(op);
            phi_outer->add_operand(bb);
        }
        inner.emplace_back(phi_outer, preheader);
        for (unsigned i = 0; i < phi->get_num_operand(); ++i)
            if (phi->get_operand(i))
                phi->get_operand(i)->remove_use(phi, i);
        phi->remove_all_operands();

        for (auto [op, bb] : inner) {
            assert(op->get_type() == phi->get_type() and
                   "op_type must equal to phi_type");
            phi->add_operand(op);
            phi->add_operand(bb);
        }
    }

    // connect
    auto in_bbs = header->get_pre_basic_blocks();
    for (auto in_bb : in_bbs) {
        if (loop.get_back_edges_nodes().find(in_bb) ==
            loop.get_back_edges_nodes().end()) {
            // connect in_bb to preheader
            in_bb->get_terminator()->replace_operand(header, preheader);
            assert(find(in_bb->get_succ_basic_blocks().begin(),
                        in_bb->get_succ_basic_blocks().end(),
                        header) == in_bb->get_succ_basic_blocks().end());
        }
    }
    BranchInst::create_br(header, preheader);

    return preheader;
}
/*Loop exit-block insertion guarantees that all exit blocks from the loop
 * (blocks which are outside of the loop that have predecessors inside of the
 * loop) only have predecessors from inside of the loop (and are thus dominated
 * by the loop header). This simplifies transformations such as store-sinking
 * that are built into LICM.
 * 保证循环的退出块的后继的前驱只有循环内的bb*/
void LoopSimplify::exit_insertion(BasicBlock *exiting, BasicBlock *outer_bb) {
    auto func = exiting->get_parent();
    auto exit = BasicBlock::create(func->get_parent(), "", func);
    exiting->get_terminator()->replace_operand(outer_bb, exit);
    BranchInst::create_br(outer_bb, exit);

    // phi
    for (auto &&inst : outer_bb->get_instructions()) {
        if (not inst.is_phi()) {
            break;
        }
        auto phi = dynamic_cast<PhiInst *>(&inst);
        for (size_t i = 0; i < phi->get_num_operand(); i++) {
            if (phi->get_operand(i) == exiting) {
                phi->set_operand(i, exit);
            }
        }
    }
}

void LoopSimplify::run(Function *func, AnalysisPassManager &APM) {

    if (func->is_declaration()) {
        return;
    }
    auto &&loops = APM.getResult<LoopDetection>(func)->get_loops();
    for (auto &&loop : loops) {
        if (loop->get_preheader() == nullptr) {
            preheader_insertion(*loop);
        }
        for (auto [exiting, exit] : loop->get_exits()) {
            if (exit == nullptr) {
                auto &bbs_in_loop = loop->get_blocks();
                auto exit_target = find_if(
                    exiting->get_succ_basic_blocks().begin(),
                    exiting->get_succ_basic_blocks().end(),
                    [&](BasicBlock *bb) {
                        return not(find(bbs_in_loop.begin(), bbs_in_loop.end(),
                                        bb) != bbs_in_loop.end());
                    });
                assert(exit_target != exiting->get_succ_basic_blocks().end());
                exit_insertion(exiting, *exit_target);
            }
        }
        // check LoopSimplify
        assert(loop->isLoopSimplifyForm() and
               "应该是循环多个回边的情况没有处理");
    }
    APM.invalidateAll<Function>(func);
}
